/* The foo_precompiled.h header is included by a /FI option when this
   source is used to create a precompiled header.  Include it here
   explicitly to allow dependency scanning to detect the dependency
   whether or not the include path is known to the scanner.  */
#include "include/foo_precompiled.h"
