MODULE TEST_MODULE
  INTERFACE
     FUNCTION TEST_MODULE_FUNCTION(A,B)
       REAL :: TEST_MODULE_FUNCTION
       REAL, INTENT(IN) :: A
       REAL, INTENT(IN) :: B
     END FUNCTION TEST_MODULE_FUNCTION
  END INTERFACE
END MODULE TEST_MODULE
